/* Common.c -- Copyright 1989, 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/*
 * $Id: Common.c,v 1.6 93/11/28 15:15:03 lee Exp $
 *
 * Handle Common Word file
 *
 */

#include "error.h"

#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include "globals.h" /* defines and declarations for database filenames */
#include "fileinfo.h"
#include "wordinfo.h"
#include "wordrules.h"
#include "emalloc.h"

/** Unix system calls that need to be declared: **/
    /* (none) */
/** C Library functions that nees to be declared: **/

/** lqtext functions that need to be declared: **/

/** Functions from this file that need to be declared: **/

API void InsertCommonWord();

/** **/

/** **/

extern int AsciiTrace;

typedef struct s_WordList {
    char *Word;
    struct s_WordList *Next;
} t_WordList;

static t_WordList *CommonWordsLow = 0;
static t_WordList *CommonWordsHigh = 0;
static char FirstCharBitMap[16] = {
    0,
};

API int
TooCommon(WordInfo)
    t_WordInfo *WordInfo;
{
    register char *Word = WordInfo->Word;
    register t_WordList **WP;

    if ((FirstCharBitMap[WordInfo->Word[0] >> 5] &
				(01 << (WordInfo->Word[0] & 07))) == 0) {
	return 0;
    }

    WP = (WordInfo->Word[0] < 'm') ? &CommonWordsLow : &CommonWordsHigh;

    for (; *WP; WP = &(*WP)->Next) {
	int i = STRCMP((*WP)->Word, Word);

	if (i == 0) return 1; /* yes, it's common */
	else if (i > 0) return 0;
    }
    return 0;
}

/*TODO: use fReadWord() */
API int
ReadCommonWords(CommonFile)
    char *CommonFile;
{
    extern char *fgets();
    extern int AsciiTrace;
    static void InsertqxxWords();

    extern FILE *fEopen();
    FILE *fd;
    char Buffer[200];
    t_WordInfo W;
    char *Root;
    t_WordList *WP;

    fd = fEopen(E_FATAL, CommonFile, "common word list", "r");

    while (fgets(Buffer, sizeof(Buffer), fd) != (char *) 0) {
	register char *p;
	char *Start;

	for (p = Buffer; *p; p++) {
	    if (*p == '#') break;
	    if (StartsWord(*p)) break;
	}

	if (*p == '#' || !*p) {
	    continue;
	}

	Start = p;

	for (; *p; p++) {
	    if (!EndsWord(*p)) {
		if (OnlyWithinWord(*p) && EndsWord(p[1])) {
		    continue;
		}
		break;
	    }
	}

	if (p - Start + 1 < MinWordLength) continue;

	*p = '\0'; /* delete trailing \n or whatever */
	W.Word = Start;
	W.Length = p - Start; /* length excludes the \0 */

	Root = WORDROOT(&W);
	InsertCommonWord(Root);
    }
    (void) fclose(fd);

    if (AsciiTrace > 2) {
	for (WP = CommonWordsLow; WP; WP = WP->Next) {
	    fprintf(stderr, "%s: Ignore \"%s\"\n", CommonFile, WP->Word);
	}
	for (WP = CommonWordsHigh; WP; WP = WP->Next) {
	    fprintf(stderr, "%s: Ignore \"%s\"\n", CommonFile, WP->Word);
	}
    }

    /* insert words of the form qxx* automatically */
    InsertqxxWords();

    return 0;
}

static void
InsertqxxWords()
{
    char *Buffer;
    register char *p;
    
    Buffer = emalloc("qxx buffer", MaxWordLength + 1);

    Buffer[0] = 'q';
    for (p = &Buffer[1]; p - Buffer < MaxWordLength; p++) {
	*p = 'x';
	p[1] = '\0';
	InsertCommonWord(Buffer);
    }

    efree(Buffer);
}

API void
InsertCommonWord(Root)
    char *Root;
{
    register t_WordList **WP;
    t_WordList *W;

    WP = (Root[0] < 'm') ? &CommonWordsLow : &CommonWordsHigh;
    for (; *WP; WP = &(*WP)->Next) {
	int i = STRCMP((*WP)->Word, Root);

	if (i == 0) return;
	else if (i > 0) break;
    }
    /* insert it before this one! */
    W = (*WP);
    (*WP) = (t_WordList *) emalloc("Common Word Entry", sizeof(t_WordList));
    (*WP)->Word = emalloc("Common Word", strlen(Root) + 1);
    (void) strcpy((*WP)->Word, Root);
    (*WP)->Next = W;

    /* set the corresponding bit in the bitmap */
    FirstCharBitMap[Root[0] >> 5] |= (01 << (Root[0] & 07));
    return;
}
