/* Defaults.c -- Copyright 1989,1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 *
 * $Id: Defaults.c,v 1.18 93/11/28 15:12:41 lee Exp $
 *
 */

#define DefineThem /* turn externs off so we do initialisations here */
# include "globals.h" /* defines and declarations for database filenames */
#undef DefineThem
#undef EXTERN

#include "error.h"

#include <fcntl.h>
#include <errno.h>

#ifdef SYSV
extern int _filbuf(); /* this must appear before stdio.h is included... */
#endif
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include "emalloc.h"
#include <sys/types.h>

#ifdef HAVE_GETPWUID
# include <pwd.h>
#endif

#include "emalloc.h"
#include "fileinfo.h"
#include "wordinfo.h"
#include "phrase.h"

/* System and Library calls used in this function:
 *
 */
extern int open(), close();
extern void exit();

extern int atoi(), strcmp(), strlen();
#ifndef tolower
 extern int tolower();
#endif
extern char *strcpy();
extern int cknatstr();

PRIVATE int NextChar();
PRIVATE void ReadDefaultFile();
PRIVATE char *FromWhereString();

/** API functions used here **/
extern int MakeDocPath(); /* hand it DOCPATH... */
extern int ReadCommonWords(), IsDir();
extern int cknatstr();
extern char *joinstr3();
extern char *mkdbm();
char *joinstr3();
extern void lqGetFileModes();

/** **/

typedef enum {
    FW_Cmdline,
    FW_Envvar,
    FW_Default, /* use the default */
    FW_File, /* from the config file */
    FW_None /* don't use any at all */
} t_FromWhere;

static t_FromWhere DirFromWhere = FW_Default;
static t_FromWhere CommonFromWhere = FW_Default;
static t_FromWhere DocFromWhere = FW_Default;

extern int AsciiTrace;


/* should PCM_HalfCase be in globals.h??? */
t_PhraseCaseMatch PhraseMatchLevel = PCM_HalfCase;

API void
SetDefaults(argc, argv)
    int argc;
    char **argv;
{
    extern char *getenv();
    extern char *progname;
    char *p;

    /* main() should have set progname.  If it didn't. we don't strip
     * the leading / as this is (I hope!) a testing and not a production
     * version... and an early test at that!
     */
    if (!progname || !*progname) progname = argv[0];

    /* loop over arguments, looking for
     * -d -- set directory for database
     * -c -- common words file
     *
     * don't use getopts, as we'll be using that later in main(),
     * and it doesn't like being called twice.
     * As a result, main() should ignore the z: option.
     */
    while (--argc > 0) {
	if (**++argv == '-' || **argv == '+') {
	    char TurnOn = (**argv == '-');

	    switch((*argv)[1]) {
	    case 'm': /* precise matching */
		argv[0][1] = 'z'; /* so it gets ignored by getopt */

		if (!*(p = &argv[0][2])) {
		    if (argc > 1) {
			argc--; argv++;
			p = (*argv);
		    } else {
			Error(E_FATAL|E_XHINT,
			    "-m must be followed by a, h or p"
			);
		    }
		}
		if (p[1]) {
		    Error(E_FATAL|E_XHINT,
			"-m must be followed by a, h or p, not \"%s\"",
			p
		    );
		}

		switch (*p) {
		case 'p': /* precise */
		    PhraseMatchLevel = PCM_SameCase;
		    break;
		case 'h': /* heuristic */
		    PhraseMatchLevel = PCM_HalfCase;
		    break;
		case 'a': /* any, approxmate */
		    PhraseMatchLevel = PCM_AnyCase;
		    break;
		default:
		    Error(E_FATAL|E_XHINT, "-m must be followed by p, h or a");
		}
		break;

	    case 'v': /* -v is the same as -t1 */
		argv[0][1] = 'Z'; /* so it gets ignored by getopt */
		++AsciiTrace;
		break;
	    case 't': /* trace level */
		argv[0][1] = 'z'; /* so it gets ignored by getopt */
		if (argv[0][2] != '\0') {
		    p = &argv[0][2];
		} else {
		    if (argc > 1) {
			argc--;
			p = (*++argv);
		    } else {
			p = "1";
		    }
		}
		if (cknatstr(p)) {
		    AsciiTrace = atoi(p);
		} else {
		    Error(E_FATAL|E_XHINT,
			"-t must be followed by a whole number, not \"%s\"",
			p
		    );
		}

		if (AsciiTrace <= 0) AsciiTrace = 1;

		Error(E_WARN,
		    "-t: trace level set to %d",
		    AsciiTrace
		);
		break;
	    case 'c': /* common file */
		if (TurnOn) {
		    CommonFromWhere = FW_Cmdline;
		    argv[0][1] = 'z'; /* so it gets ignored by getopt */
		    if ((*argv)[2] != '\0') {
			CommonWordFile = &(*argv[2]);
		    } else {
			if (argc > 1) {
			    CommonWordFile = argv[1];
			    argc--; argv++;
			} else {
			    Error(E_FATAL|E_XHINT,
				"-c option must be followed by a filename"
			    );
			}
		    }
		} else { /* Turn off, +c, may be undocumented right now */
		    CommonFromWhere = FW_None;
		    argv[0][1] = 'z'; /* so it gets ignored by getopt */
		    argv[0][0] = '-'; /* don't confuse getopt */
		    break;
		}
		break;
	    case 'd':
		argv[0][1] = 'z'; /* so it gets ignored by getopt */
		DirFromWhere = FW_Cmdline;
		if (argv[0][2] != '\0') {
		    DatabaseDir = &argv[0][2];
		} else {
		    if (argc > 1) {
			DatabaseDir = argv[1];
			argc--; argv++;
		    } else {
			Error(E_FATAL|E_XHINT,
			    "%cd must be followed by a directory name",
			    TurnOn ? '-' : '+'
			);
		    }
		}
		break;
	    } /* end switch */
	} else {
	    /* not an option, so stop looking */
	    break;
	}
    } /* end while */

    /* now we have parsed the command line arguments, so look for the
     * default directory
     */
    if (DirFromWhere == FW_Default) {
	char *t;

	if ((t = getenv("LQTEXTDIR")) != (char *) 0) {
	    DatabaseDir = emalloc("database directory name", strlen(t) + 1);
	    (void) strcpy(DatabaseDir, t);
	    DirFromWhere = FW_Envvar;
	} else {
#ifdef UNDERHOME
	    char *home = getenv("HOME");

# ifdef HAVE_GETPWUID /* POSIX and BSD only, plus maybe NFS or NIS systems */

	    if (!home) {

# ifndef HAVE_UID_T
#  ifdef SYSV
		typedef unsigned short uid_t;
#  else
		typedef unsigned int uid_t;
#  endif
# endif /*!HAVE_UID_T*/

		extern uid_t getuid();
		extern struct passwd *getpwuid(/* uid_t uid */);
		struct passwd *pw;

		uid_t Me = getuid();

		pw = getpwuid(Me);

		if (pw && pw->pw_dir && pw->pw_dir[0] != '\0') {
		    home = pw->pw_dir;
		}
	    }

# endif /* HAVE_GETPWNAM */

	    if (home) {
		DatabaseDir = joinstr3(home, "/", UNDERHOME);
		if (!IsDir(DatabaseDir)) {
		    Error(E_FATAL|E_XHINT|E_SYS,
			"default database directory \"%s\" inaccessible",
			DatabaseDir
		    );
		}
	    } else {
		Error(E_FATAL, "can't find your login directory ($HOME)");
	    }
#endif /* UNDERHOME*/
	    /* in either case it's the default... */
	    DirFromWhere = FW_Default;
	}
    }

    if (!DatabaseDir || !*DatabaseDir) {
	/* This can happen if there is no default, or if the user types
	 * lqword -d ""
	 * just to be malicious :-)
	 */
	Error(E_FATAL|E_XHINT|E_USAGE,
	    "You must give a database directory with -d or $LQTEXTDIR"
	);
    }

    /* IsDir is in DocPath.c -- perhaps this should be, too. */
    if (!IsDir(DatabaseDir)) {
	Error(E_FATAL,
	    "lq-text database \"%s\"%s is not a directory",
	    DatabaseDir,
	    FromWhereString(DirFromWhere)
	);
    }

    /* set default filenames */
#define IfNot(x, y) ((x) ? (x) : (y))

    FileIndex = mkdbm(DatabaseDir, IfNot(FileIndex, FILEINDEX));
    WordIndex = mkdbm(DatabaseDir, IfNot(WordIndex, WORDINDEX));

    DataBase = joinstr3(DatabaseDir, "/", IfNot(DataBase, DATABASE));
    FreeFileName =
	    joinstr3(DatabaseDir, "/", IfNot(FreeFileName, FREELISTFILE));
    WidIndexFile =
	    joinstr3(DatabaseDir, "/", IfNot(WidIndexFile, WIDINDEXFILE));

    if (CommonFromWhere == FW_Default) {
	char *c = getenv("LQCOMMON");

	if (c) {
	    CommonWordFile = emalloc("Common Word file name", strlen(c) + 1);
	    (void) strcpy(CommonWordFile, c);
	    CommonFromWhere = FW_Envvar;
	}
    }

    ReadDefaultFile();

    if (AsciiTrace) {
	fprintf(stderr, "%s: lqtext directory \"%s\"\n",progname,DatabaseDir);
    }

    if (CommonFromWhere != FW_None && CommonWordFile && *CommonWordFile) {
	int c;

	if (*CommonWordFile != '/') {
	    CommonWordFile = joinstr3(DatabaseDir, "/", CommonWordFile);
	}

	c = Eopen(E_FATAL|E_SYS,
	    CommonWordFile,
	    "common word file",
	    O_RDONLY,
	    0
	);
	
	(void) close(c); /* it's OK */
    }

    if ((p = getenv("DOCPATH")) != (char *) 0) {
	switch (DocFromWhere) {
	case FW_File:
	    if (AsciiTrace > 1) {
		Error(E_WARN, "DOCPATH (%s) overrides %s (%s)",
#ifdef CONFIGFILE
				p, CONFIGFILE, DocPath
#else
				p, "README", DocPath
#endif
		);
	    }
	    efree(DocPath);
	    /* FALL THROUGH */
	case FW_Default:
	default: /* ? */
	    DocPath = emalloc("DOCPATH", (unsigned) (strlen(p) + 1));
	    (void) strcpy(DocPath, p);
	    DocFromWhere = FW_Envvar;
	    break;
	}
    }

    if (!DocPath || !*DocPath) {
	DocPath = ".";
    }

#define SetOrNot(s) ( (s && *s) ? s : (s ? "[empty]" : "[null]" ) )

    /* this is always here -- it's only checked once, and is actually
     * rather useful.
     */
    if (AsciiTrace > 2) {
	fprintf(stderr, "%s: CommonWordFile = \"%s\"\n", progname,
		SetOrNot(CommonWordFile));
	fprintf(stderr, "%s: DatabaseDir = \"%s\"\n", progname,
		SetOrNot(DatabaseDir));
	fprintf(stderr, "%s: DocPath  = \"%s\"\n", progname,
		SetOrNot(DocPath));
	fprintf(stderr, "%s: FileIndex = \"%s\"\n", progname,
		SetOrNot(FileIndex));
	fprintf(stderr, "%s: WordIndex = \"%s\"\n", progname,
		SetOrNot(WordIndex));
	fprintf(stderr, "%s: DataBase = \"%s\"\n", progname,
		SetOrNot(DataBase));
	fprintf(stderr, "%s: WidIndexFile = \"%s\"\n", progname,
		SetOrNot(WidIndexFile));
    }

    (void) MakeDocPath(DocPath);
    /* DocPath is no longer needed, so getenv() can be called again now */

    if (CommonFromWhere != FW_None && CommonWordFile && *CommonWordFile) {
	(void) ReadCommonWords(CommonWordFile);
    } else {
	(void) ReadCommonWords("/dev/null");
    }
}

API void
DefaultUsage()
{
    fprintf(stderr, "\
	-c file	-- ignore words that are listed in the namd file\n\
	-d dir	-- use the lq-text database in the named directory\n\
	-m c	-- set matching criteria -- c is \"p\", \"h\" or \"a\"\n");
    if (AsciiTrace) {
	fprintf(stderr, "\
		   -m p  uses precise matching, where CaSe is significant;\n\
		   -m h  uses heuristic matching, which is the default, and\n\
		   -m a  uses approximate matching.\n");
    }

    fprintf(stderr, "\n\
	-t N	-- set trace level t N (default is zero)\n\
	-x	-- print %s explanation\n\
	-xv	-- print %s explanation\n\
	-V	-- print version information\n\
	-v	-- be verbose (same as -t 1)\n",
			AsciiTrace ? "a shorter" : "this",
			AsciiTrace ? "this" : "a longer");
    if (AsciiTrace) {
	fprintf(stderr, "\
The current database directory is \"%s\";\n\
%s will search the path \"%s\" for documents.\n", DatabaseDir, progname, DocPath);
    }
}

PRIVATE char *
FromWhereString(Whence)
    t_FromWhere Whence;
{
    char *msg = "";

    switch (Whence) {
    case FW_Cmdline:
	msg = " (specified with the -d option)";
	break;
    case FW_Envvar:
	msg = " (from $LQTEXTDIR)";
	break;
    case FW_Default:
	msg = " (compiled-in default)";
	break;
    case FW_File:
	{
#ifdef CONFIGFILE
	char *ReadMe = joinstr3(DatabaseDir, "/", CONFIGFILE);
#else
	char *ReadMe = joinstr3(DatabaseDir, "/", "README");
#endif
	    char *fmt = " (from the config file, \"%s\")";
	    msg = emalloc("message for FromWhere", strlen(fmt) + strlen(ReadMe));
	    if (!msg) {
		msg = " from the config file";
		break;
	    }
	    (void) sprintf(msg, fmt, ReadMe);
	    (void) efree(ReadMe);
	}
	break;
    case FW_None:
	msg = " (disabled by + option)";
	break;
    default:
	break;
    }
    return msg;
}

#define LCNOMAP 0 /* Token -- leave case alone */
#define LCMAP 1  /* map to lower case */

static int RMLine = 0;

static char *Bool_Yes[] = {
    "on",
    "yes",
    "true",
    "#t", /* for scheme */
    "t",
    "1",
    0
};

static char *Bool_No[] = {
    "off",
    "no",
    "false",
    "#f",
    "()",
    "nil",
    "0",
    0
};

PRIVATE int
StringToBool(String)
    char *String;
{
    char *tmp = emalloc("StringToBool", strlen(String) + 1);
    register char *p, *q;
    int i;

    p = String;
    q = tmp;
    while (*p) {
	if (isascii(*p) && isupper(*p)) {
	    *q = tolower(*p);
	} else {
	    *q = *p;
	}
	q++; p++;
    }
    *q = '\0';

    for (i = 0; Bool_Yes[i] != (char *) 0; i++) {
	if (STREQ(Bool_Yes[i], tmp)) {
	    (void) efree(tmp);
	    return 1;
	}
    }

    for (i = 0; Bool_No[i] != (char *) 0; i++) {
	if (STREQ(Bool_No[i], tmp)) {
	    (void) efree(tmp);
	    return 0;
	}
    }

    return -1;
}

PRIVATE void
ReadDefaultFile()
{
    extern int errno;

    static char *NextToken(); /* see below */
#ifdef CONFIGFILE
    char *ReadMe = joinstr3(DatabaseDir, "/", CONFIGFILE);
#else
    char *ReadMe = joinstr3(DatabaseDir, "/", "README");
#endif
    FILE *fp;
    char *Token;


    /* This is paranoid... */
    if (!ReadMe || !*ReadMe) {
	Error(E_FATAL|E_INTERNAL|E_BUG,
	    "%s: %d: Configuration filename ended up %s",
	    __FILE__, __LINE__,
	    SetOrNot(ReadMe)
	);
    }

    errno = 0;
    if ((fp = fopen(ReadMe, "r")) == (FILE *) 0) {
	if (errno == EPERM) {
	    Error(E_FATAL|E_SYS,
		"you don't have permission to read \"%s\"",
		ReadMe
	    );
	} else {
	    Error(E_FATAL|E_SYS,
		"can't open configuration file \"%s\"",
		ReadMe
	    );
	}
	/*NOTREACHED*/
	exit(1);
    }

    /* Read README up to an "end" line, ignoring lines starting with # */

    while ((Token = NextToken(fp, ReadMe, LCMAP)) != (char *) 0) {
	if (STREQ(Token, "end")) goto finish;
	if (STREQ(Token, "common")) {
	    if (!(Token = NextToken(fp, ReadMe, LCNOMAP))) {
		Error(E_FATAL,
		    "%s %d: unexpected eof at common file\n",
		    ReadMe,
		    RMLine
		);
	    } else if (CommonFromWhere == FW_Default) {
		CommonWordFile =
		    emalloc("CommonWordFilename", (strlen(Token) + 1));
		(void) strcpy(CommonWordFile, Token);
		CommonFromWhere = FW_File;
	    }
	} else if (STREQ(Token, "echo")) {
	    char *p = NextToken(fp, ReadMe, LCNOMAP);

	    if (!p) p = "";
	    fprintf(stderr, "%s: %s\n", progname, p);
	} else if (STREQ(Token, "minwordlength") || STREQ(Token, "minlen")) {
	    if (!(Token = NextToken(fp, ReadMe, LCNOMAP))) {
		Error(E_FATAL,
		    "%s: %d: unexpected eof at common file\n",
		    ReadMe,
		    RMLine
		);
	    } else if (cknatstr(Token)) {
		MinWordLength = atoi(Token);
	    } else {
		Error(E_FATAL,
		    "config file %s: %d: minwordlength must be followed by a number.",
		    ReadMe,
		    RMLine
		);
	    }
	} else if (STREQ(Token, "wordlist") || STREQ(Token, "wordsinindex")) {
	    if (!(Token = NextToken(fp, ReadMe, LCNOMAP))) {
		Error(E_FATAL,
		    "Config file %s: %d: expected on or off",
		    ReadMe,
		    RMLine
		);
	    } else {
		switch (StringToBool(Token)) {
		case 0:
		    WordsInWordIndex = 0;
		    break;
		case 1:
		    WordsInWordIndex = 1;
		    break;
		case -1:
		    Error(E_FATAL,
"Config file %s: %d: wordlist must be followed by on or off, not %s",
			ReadMe,
			RMLine,
			Token
		    );
		}
	    }
	} else if (STREQ(Token, "maxwordlength") || STREQ(Token, "maxlen")) {
	    if (!(Token = NextToken(fp, ReadMe, LCNOMAP))) {
		Error(E_FATAL,
		    "%s: %d: unexpected eof at common file\n",
		    ReadMe,
		    RMLine
		);
	    } else if (cknatstr(Token)) {
		MaxWordLength = atoi(Token);
	    } else {
		Error(E_FATAL,
		    "config file %s: %d: maxwordlength must be followed by a number.",
		    ReadMe,
		    RMLine
		);
	    }
	} else if (STREQ(Token, "path") || STREQ(Token, "docpath")) {
	    if (!(Token = NextToken(fp, ReadMe, LCNOMAP))) {
		Error(E_FATAL,
		    "%s: %d: unexpected end of file in docpath line",
		    ReadMe,
		    RMLine
		);
	    } else {
		DocPath = emalloc("DocPath", (unsigned) (strlen(Token) + 1));
		(void) strcpy(DocPath, Token);
		DocFromWhere = FW_File;
	    }
	} else {
	    Error(E_FATAL,
		"\"%s\": %d: \"%s\" unexpected (use # to start a comment)",
		ReadMe,
		RMLine,
		Token
	    );
	}
    } /* while */

finish:
    if (WordsInWordIndex && (MaxWordLength + 10 > WIDBLOCKSIZE)) {
	Error(E_FATAL,
	    "Config file %s: Max word length %d > WIDBLOCKIZE (%d) + 10",
	    ReadMe,
	    MaxWordLength,
	    WIDBLOCKSIZE
	);
    }
    (void) fclose(fp);
    return;
}

PRIVATE char *
NextToken(fd, Name, Map)
    FILE *fd;
    char *Name;
    int Map;
{
    int ch;
    static char *buf = (char *) NULL;
    static int BytesAllocated = 0;
    register char *q = buf;
    int InQuote = 0;
    int OriginalMap = Map;

    if (!buf || !BytesAllocated) {
	BytesAllocated = 50;
	buf = emalloc("NextToken line buffer", BytesAllocated);
    }

    q = buf;

    while ((ch = NextChar(fd, Name, Map)) != EOF) {

	/* Ensure that we don't fall off the end of the buffer */
	if (q - buf >= BytesAllocated) {
	    int WhereWeWere = q - buf;

	    BytesAllocated += 50;
	    buf = erealloc(buf, BytesAllocated);
	    q = &buf[WhereWeWere];
	}

	switch (ch) {
	case '"': case '\'':
	    if (!InQuote) {
		if (q == buf) { /* start a quoted string */
		    InQuote = ch;
		    Map = 0; /* no case conversion inside strings */
		} else {
		    Error(E_FATAL,
			"%s: %d: character (%c) must be quoted",
			Name, RMLine,
			ch
		    );
		}
	    } else {
		if (ch == InQuote) { /* end of the string */
		    *q = '\0';
		    return buf;
		} else { /* nested quote: "....'..." */
		    *q++ = ch;
		}
	    }
	    break;
	case '\\':
	    /* swallow the backslash: */
	    if ((ch = NextChar(fd, Name, Map)) == EOF) {
		Error(E_FATAL,
		    "%s; %d: EOF after \\ unexpected!",
		    Name,
		    RMLine
		);
	    }
	    *q++ = ch;
	    break;
	case ' ':
	    if (InQuote) {
		*q++ = ch;
		break;
	    }
	    /* FALL THROUGH */
	case '\n':
	    if (InQuote) {
		*q = '\0';
		Error(E_FATAL,
		    "%s: %s: missing quote after %c%s",
		    Name, RMLine,
		    InQuote, buf
		);
	    }
	    if (q - buf >= BytesAllocated) {
		int WhereWeWere = q - buf;

		BytesAllocated += 50;
		buf = erealloc(buf, BytesAllocated);
		q = &buf[WhereWeWere];
	    }
	    *q = '\0';
	    if (q > buf) return buf;
	    else return NextToken(fd, Name, OriginalMap);
	    /*NOTREACHED*/
	    break;
	default:
	    *q++ = ch;
	    break;
	}
    }
    if (q > buf) {
	Error(E_FATAL, "%s: %d: unexpected end of file", Name, RMLine);
    }
    (void) efree(buf);
    buf = 0;
    BytesAllocated = 0;
    return (char *) 0;
}

PRIVATE int
NextChar(fd, Name, Map)
    FILE *fd;
    char *Name;
    int Map;
{
    int ch;

    while ((ch = getc(fd)) != EOF) {
	switch (ch) {
	case '#':
	    do {
		if ((ch = getc(fd)) == EOF) {
		    Error(E_FATAL,
			"%s: %d: unexpected end of file inside comment",
			Name,
			RMLine
		    );
		}
	    } while (ch != '\n');
	    /* ASSERT: ch == '\n' */
	    ++RMLine;
	    break;
	case '\n':
	    ++RMLine; 
	    if (!Map) return ch;
	    /* else FALL THROUGH */
	case ' ': case '\t': case '\f': case '\r':
	    if (!Map) {
		return ' ';
	    }

	default:
	    return (Map && isupper(ch)) ? tolower(ch) : ch;
	}
    } /* while */
    return EOF;
}
