/* lqaddfile.c -- Copyright 1989, 1990 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* addfile -- add a file to the LQ-Text text retrieval index
 * Liam Quin, August 1989 and later...
 *
 * $Id: lqaddfile.c,v 1.29 93/12/08 23:16:45 lee Exp $ 
 */

static char *Version = "@(#) $Id: lqaddfile.c,v 1.29 93/12/08 23:16:45 lee Exp $";

#ifdef SYSV
extern int _filbuf(); /* used but not defined in stdio.h */
#endif
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef BSD
# include <strings.h>
#else
# include <string.h>
#endif

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include "fileinfo.h"
#include "wordinfo.h"
#include "wordrules.h"
#include "filter.h"
#include "emalloc.h"
#include "addfile.h"
#include "Revision.h"

#include "readfile.h"

#ifdef SYSV
#define TOLOWER(ch) ch = tolower(ch)
#else
#define TOLOWER(ch) if (isupper(ch)) ch = tolower(ch)
#endif

void DestroyFileInfo();
extern int SaveFileInfo();
void AddStream(), AddFrom();
extern lqWriteAccess(); /* Allow write access to the database */
/* Symbol Table Interface */
extern void AddWord(), WriteCurrentMaxWID();
extern void DumpCache(), cleanupdb();
extern int GetFilterType();
extern void SetDumpThresh();
int AddFile();

/** System calls and library routines used in this file: **/
/** System calls: **/
extern void exit();
extern int stat();
/** Library Functions: **/
extern int atoi();
#ifndef tolower
 extern int tolower();
#endif
extern void perror();

/** liblqtext functions: **/
extern t_WordInfo *FReadWord();

/** Functions within this file that need declaring: **/
static void UnIndex();
/**/

char *progname = "@(#) $Id: lqaddfile.c,v 1.29 93/12/08 23:16:45 lee Exp $";

#ifdef USE_LINENUMBERS
static int UseLineNumbers = 0;
#endif

extern int AsciiTrace; /* provide increasingly verbose info if not zero */

static int MightBeUpdates = 0;

static int SignalFlag = 0;

int
SignalHandler()
{
    extern void FlushCache();

    ++SignalFlag;
    if (SignalFlag > 3) {
	FlushCache(0);
	WriteCurrentMaxWID();
	cleanupdb(); /* empty the dbm cache */
	Error(E_FATAL,
	    "received %d signals to quit, exiting; db may be corrupt!.",
	    SignalFlag
	);
    }
    return 0;
}

int
main(argc, argv)
    int argc;
    char *argv[];
{
    extern void SetHashSize();
    extern char *strrchr();
    extern int getopt(), cknatstr();
    extern void SetDefaults();
    extern char *optarg;
    extern int optind;
    extern int MaxWordsInCache; /* see wordtable.c */

    int c;
    int ErrorFlag = 0;
    int DoNothing = 0;
    char *InputFile = (char *) 0;

#ifdef MALLOCTRACE
    malloc_debug(2);
#endif

    progname = argv[0]; /* retain the full path at first */

#ifdef M_MXFAST
    /* The new symbol table doesn't allocate large numbers of t_WordPlace,
     * so we want this number small...
     */
# if 0
    (void) mallopt(M_MXFAST, sizeof(t_WordPlace));
# else
    (void) mallopt(M_MXFAST, 6); /* i.e. typical word length with \0 */
# endif
    /* may need to comment mallopt() out entirely for BSD -- use ifndef.
     * seems to work under SunOS, though.
     * When it works, it says "Allocate 100 or so chunks of this size at a
     * time, and whenver I ask for this much or less, give me one of the
     * chunks".  Clearly it had better not be too large, but it is a big
     * win with a structure allocated for every occurrence of every word!
     */
#endif
    SetDefaults(argc, argv);

    while ((c = getopt(argc, argv, "w:f:H:M:xVuZz:")) != -1) {
	switch (c) {
	case 'M':
	    if (!cknatstr(optarg)) {
		Error(E_FATAL|E_USAGE|E_XHINT,
		    "-M must be given a number >= 0, not \"%s\"",
		    optarg
		);
	    }
	    SetDumpThresh(atoi(optarg));
	    break;
	case 'H':
	    if (!cknatstr(optarg)) {
		Error(E_FATAL|E_USAGE|E_XHINT,
		    "-H must be given a hash table size >= 1, not \"%s\"",
		    optarg
		);
	    }
	    SetHashSize(atoi(optarg));
	    break;
	case 'u':
	    ++MightBeUpdates;
	    break;
	case 'w':
	    if (!cknatstr(optarg)) {
		Error(E_FATAL|E_USAGE|E_XHINT,
		    "-w must be given a number >= 0, not \"%s\"",
		    optarg
		);
	    }
	    MaxWordsInCache = atoi(optarg);
	    break;
	case 'Z':
	case 'z':
	    break; /* work done in SetDefault() */
	case 'V':
	    fprintf(stderr, "%s: Release: %s\n", LQTEXTREVISION);
	    fprintf(stderr, "%s: Revision: %s\n", progname, Version);
	    DoNothing = 1;
	    break;
	case 'f':
	    if (InputFile) {
		Error(E_USAGE|E_XHINT|E_FATAL,
		    "only one -f option allowed; use -xv for explanation"
		);
	    }
	    InputFile = optarg;
	    break;
	case 'x':
	    ErrorFlag = (-1);
	    break;
	default:
	case '?':
	    ErrorFlag = 1;
	}
    }

    if ((progname = strrchr(progname, '/')) != (char *) NULL) {
	++progname; /* step over the last / */
    } else {
	progname = argv[0];
    }

    if (ErrorFlag > 0) {
	fprintf(stderr, "use %s -x or %s -xv for an explanation.\n",
							progname, progname);
	exit(1);
    } else if (ErrorFlag < 0) { /* -x was used */
	fprintf(stderr, "%s -- add files to an lq-text retrieval database\n",
								    progname);

	fputs("Options are:\n\
	-f file -- read the list of files to index from \"file\"\n\
	-c file	-- file contains a list of common words to be ignored\n\
	-d dir	-- use the lq-text database in the named directory\n\
	-M n	-- try not to flush cache entries with n or more entries\n\
	-t N	-- set the trace level to N [default: N = 0]\n\
	-V	-- print Version number and exit\n\
	-v	-- be verbose\n\
	-w n	-- dump the word-cache every n words\n\
	-x	-- print an eXplanation and exit\n\
	-xv	-- print a more detailed eXplanation and exit\n\
	--	-- all following arguments are file names\n\
\n\
", stderr);
	if (AsciiTrace == 1) {
	    /* used -v or -t1 */
	fprintf(stderr, "\n\
    Any remaining arguments are taken to be file names.  The current\n\
DOCPATH (%s) is searched for the files, and they are read and added\n\
to the index.  (If you use the -f option, you should not give filename\n\
arguments on the command line, although you can use \"-f -\" to read the\n\
list of files from standard input, one per line.\n\
Setting (with -w) the size of the cache may dramatically\n\
improve performance.  Systems with memory larger than the data can try -w0.\n\
See lqtext(1) for more information.\n", DocPath);
	}
    exit(0);

    }


#ifdef WIDINBLOCK
    {
	char *S = "********";
# ifdef ASCIITRACE
	/* remind people to recompile... */
	Error(E_WARN,
	    "%s\n%s Compiled with -DWIDINBLOCK for debugging\n%s\n",
	    S,
	    S,
	    S
	);
# else
	/* don't allow -DWIDINBLOCK without -DASCIITRACE */
	Error(E_BUG, "Compiled with -DWIDINBLOCK but not -DASCIITRACE!");
	syntax error; this prevents compilation here;
# endif /* ASCIITRACE */
    }
#endif

    if (DoNothing) {
	if (optind < argc) {
	    Error(
		E_WARN|E_XHINT,
		    "%d extra argument%s ignored...\n",
		    argc - optind,
		    argc - optind == 1 ? "" : "%s"
		);
	}
	exit(0);
    }

    lqSetSignals(SignalHandler);
    lqWriteAccess();

    if (InputFile) {
	if (optind < argc) {
	    Error(E_FATAL|E_USAGE|E_XHINT,
		"cannot give filenames after -f %s",
		InputFile
	    );
	}
	AddFrom(InputFile);
    } else for (; optind < argc; ++optind) {
	if (AddFile(argv[optind]) < 0 && AsciiTrace >= 1) {
	    if (SignalFlag) {
		Error(E_WARN,
		    "Caught signal at level %d, dumping cache",
		    SignalFlag
		);
		DumpCache(DUMP_SYNC);
		WriteCurrentMaxWID();
		cleanupdb(); /* empty the dbm cache */
		exit(1);
	    } else {
		Error(E_WARN, "%s not added to index", argv[optind]);
	    }
	}

    }

#ifndef MALLOCTRACE
    /* don't bother recaiming storage if we're about to exit, unless we
     * want to check for memory leaks afterwards.
     */
    DumpCache(DUMP_SYNC|DUMP_NOFREE);
#else
    DumpCache(DUMP_SYNC);
#endif

    WriteCurrentMaxWID();
    cleanupdb(); /* empty the dbm cache */

#ifdef MALLOCTRACE
    (void) fprintf(stderr, "%s: Malloctrace: checking...\n", progname);
    malloc_verify();
    (void) fprintf(stderr, "%s: Malloc Map\n", progname);
    mallocmap();
#endif

    exit(0);
}

void
AddFrom(Name)
    char *Name;
{
    FILE *fp;
    char *Line;

    if (Name[0] == '-' && Name[1] == '\0') {
	fp = stdin;
    } else {
	fp = fopen(Name, "r");
    }

    if (fp == (FILE *) 0) {
	Error(E_FATAL|E_SYS, "-f: can't open \"%s\"", Name);
    }

    while (fReadLine(fp, &Line, UF_NORMAL) > 0) {
	/* Note:
	 * fReadFile will silently swallow blank lines.
	 * if we use UF_NORMAL it will swallow lines that start with a #,
	 * but we don't want that here!
	 */
	if (AddFile(Line) < 0) {
	    if (SignalFlag) {
		Error(E_WARN,
		    "Caught signal at level %d -- dumping cache",
		    SignalFlag
		);
		DumpCache(DUMP_SYNC);
		WriteCurrentMaxWID();
		cleanupdb(); /* empty the dbm cache */
		exit(1);
	    } else {
		if (AsciiTrace > 1) {
		    /* AddFile should already have printed a message... */
		    Error(E_WARN, "%s not added to index", Name);
		}
	    }
	}
	(void) efree(Line);
    }

    if (fp != stdin) {
	(void) fclose(fp);
    }
}

extern int fclose(), pclose();

t_FileInfo *
MakeFileInfo(Name)
    char *Name;
{
#ifdef BSD
    extern time_t time();
#else
    extern long time();
#endif
    extern t_FID Name2FID();
    extern t_FileInfo *GetFileInfo();
    extern t_FID GetNextFID();
    FILE *MakeInput();
    struct stat StatBuf;

    t_FileInfo *FileInfo = 0;
    t_FID FID;

    if (!Name || !*Name) return (t_FileInfo *) 0; /* sanity */

    if (stat(Name, &StatBuf) < 0) {
#ifndef FindFile /* it is a macro these days... */
	extern char *FindFile();
#endif
	char *doc;

	if ((doc = FindFile(Name)) == (char *) 0) {
	    Error(E_SYS, "Can't index \"%s\"", Name);
	    return (t_FileInfo *) 0;
	}

#ifndef COMPRESS_SUPPORT
	{
	    int len = strlen(doc);

	    if (len > 3 && doc[len - 2] == '.' && doc[len - 1] == 'Z') {
		/* TODO: allow lqaddfile and filters to unpack archives etc. */
		Error(E_WARN,
		    "%s: files must be compressed after indexing",
		    doc
		);
		return (t_FileInfo *) 0;
	    }
	}
#endif /*!COMPRESS_SUPPORT*/

	if (stat(doc, &StatBuf) < 0) {
	    Error(E_SYS, "%s: can't index \"%s\"", Name, doc);
	    return (t_FileInfo *) 0;
	}
	Name = doc;
    }

    if (StatBuf.st_size == 0L) {
	if (AsciiTrace) {
	    Error(E_WARN, "%s empty -- not indexed", Name);
	}
	return (t_FileInfo *) 0;
    }
    /* See if it's in the index already: */
    if ((FID = Name2FID(Name)) != (t_FID) 0) {

	if ((FileInfo = GetFileInfo(FID)) != (t_FileInfo *) 0) {
	    /* Check to see if the file hass changed since it was last
	     * indexed.  If it has, we should delete the old one from
	     * the database and give this one a new FID, but I have
	     * not done that yet -- that's /usr/local/lib/lqtextd or
	     * something, I suppose!
	     */
	    if (FileInfo->Date >= StatBuf.st_mtime &&
				FileInfo->FileSize == StatBuf.st_size) {
		if (AsciiTrace) {
		    Error(E_WARN,
			"%s unchanged since last run -- not indexed",
			Name
		    );
		}
		DestroyFileInfo(FileInfo);
		return (t_FileInfo *) 0;
	    }
	}
    } else {
	FID = GetNextFID((long) StatBuf.st_size);
    }

    if (FileInfo == (t_FileInfo *) 0) {
	/* Allocate Structure */
	FileInfo = (t_FileInfo *) emalloc("MakeFileInfo", sizeof(t_FileInfo));

	/* Although not always necessary, call emalloc here so that a
	 * FileInfo can always be deleted with DestroyFileInfo()
	 */
	FileInfo->Name = emalloc(
	    "MakeFileInfo.Name",
	    (unsigned)(strlen(Name) + 1)
	);
	(void) strcpy(FileInfo->Name, Name);

	/* Other bits to set: */

	FileInfo->Date = StatBuf.st_mtime;
	FileInfo->FileSize = StatBuf.st_size;
	FileInfo->Stream = 0;

	/* file type */
	if ((FileInfo->FilterType = GetFilterType(FileInfo, &StatBuf)) < 0) {
	    if (AsciiTrace) {
		Error(E_WARN,
		    "%s unknown file type -- not indexed",
		    Name
		);
	    }
	    DestroyFileInfo(FileInfo);
	    return (t_FileInfo *) 0;
	}
    }

    FileInfo->FID = FID;
    FileInfo->Date = (long) time((long *) 0); /* it's a time_t on BSD */

    if ((FileInfo->Stream = MakeInput(FileInfo)) == (FILE *) 0) {
	Error(E_WARN|E_SYS,
	    "couldn't open input filter for %s -- not indexed",
	    FileInfo->Name
	);
	DestroyFileInfo(FileInfo);
	return (t_FileInfo *) 0;
    }

    return FileInfo;
}

void
DestroyFileInfo(FileInfo)
    t_FileInfo *FileInfo;
{
    if (FileInfo->Stream) {
	if (FileInfo->FilterType >= 0 &&
				FileInfo->FilterType <= MaxFilterType &&
				FilterTable[FileInfo->FilterType].close &&
				FileInfo->Stream != (FILE *) NULL) {
	    (* FilterTable[FileInfo->FilterType].close)(FileInfo->Stream);
	}
	FileInfo->Stream = (FILE *) 0;
    }
    if (FileInfo->Name) (void) efree(FileInfo->Name);
    (void) efree((char *) FileInfo);
}

int
AddFile(Name)
    char *Name;
{
    t_FileInfo *FileInfo;

    if (!Name || !*Name) return -1;

    if (MightBeUpdates) {
	char *p = index(Name, '\t');

	if (p && *p == '\t') {
	    if (p[1]) {
		*p = '\0';
		FileInfo = MakeFileInfo(Name);
		if (FileInfo == (t_FileInfo *) 0) {
		    Error(E_WARN, "old version of %s not found", p);
		} else {
		    UnIndex(FileInfo);
		    DestroyFileInfo(FileInfo);
		}
		Name = p;
	    }
	}
    }

    if ((FileInfo = MakeFileInfo(Name)) == (t_FileInfo *) 0) return -1;

    AddStream(FileInfo);
    SaveFileInfo(FileInfo);
    DestroyFileInfo(FileInfo);

    if (SignalFlag) {
	return -1;
    }

    return 0;
}

FILE *
MakeInput(FileInfo)
    t_FileInfo *FileInfo;
{
    FILE *fp;
    char *Buffer;
    unsigned BufLen;
    extern FILE *fopen(), *popen();

#define FSTRING FilterTable[FileInfo->FilterType].String

    if (FileInfo->FilterType > MaxFilterType) {
	Error(E_WARN, "filter type %d for %s too high (max %d)",
		FileInfo->FilterType, FileInfo->Name, MaxFilterType);
	return (FILE *) 0;
    }

    if (FilterTable[FileInfo->FilterType].Type != FileInfo->FilterType) {
	Error(E_FATAL|E_INTERNAL,
	    "Filter table entry %d has type %d, expected %d",
	    FileInfo->FilterType,
	    FilterTable[FileInfo->FilterType].Type,
	    FileInfo->FilterType
	);
    }

    if (FSTRING == (char *) 0) {
	return fopen(FileInfo->Name, "r");
    }

    BufLen = strlen(FileInfo->Name) * 2 + 4 + strlen(FSTRING);
	/* The +4 is to allow for an embedded " < " plus a \0;
	 * we append "< Name", but also expand %s to be the Name, hence
	 * the strlen * 2; +2 is for quotes.
	 */
    Buffer = emalloc("MakeInput command buffer", BufLen);

    (void) sprintf(Buffer, FSTRING, FileInfo->Name);
    (void) strcat(Buffer, " < ");
    {
	register char *p, *q;

	for (p = Buffer; *p; p++) {
	    /*NULLBODY*/;
	}
	*p++ = '"';
	for (q = FileInfo->Name; *q; q++) {
	    if (*q == '"' || *q == '$' || *q == '\\' || *q == '`') {
		*p++ = '\\';
	    }
	    *p++ = *q;
	}
	*p++ = '"';
	*p = '\0';
    }

    fp = popen(Buffer, "r");
    if (!fp) {
	Error(E_WARN|E_SYS, "%s: Couldn't run input command [%s]",
	    FileInfo->Name, Buffer
	);
    }
    (void) efree(Buffer);
    return fp;
}

static void
UnIndex(FileInfo)
    t_FileInfo *FileInfo;
{
    extern void DelWord();
    t_WordInfo *WordInfo;

    while ((WordInfo = FReadWord(FileInfo)) != (t_WordInfo *) 0) {
	DelWord(WordInfo, FileInfo->FID);
	if (SignalFlag) return;
    }
}

void
AddStream(FileInfo)
    t_FileInfo *FileInfo;
{
    /* I have to mark the last word in the block.
     * I do that by marking the previous word if it was in a differant block
     * than the current one.
     */
    t_WordInfo *WordInfo;
    t_WordInfo *LastWord = 0;

    if (SignalFlag > 1) return;

    while ((WordInfo = FReadWord(FileInfo)) != (t_WordInfo *) 0) {
	if (LastWord) {
	    if (LastWord->WordPlace.BlockInFile !=
	    				WordInfo->WordPlace.BlockInFile) {
		LastWord->WordPlace.Flags |= WPF_LASTINBLOCK;
	    }
	    AddWord(LastWord);
	}
	LastWord = WordInfo;
	if (SignalFlag > 1) {
	    Error(E_WARN, "Signal received during processing of %s",
		    FileInfo->Name);
	    Error(E_WARN, "That and other files may be incomplete...");
	    break;
	}
    }
    if (LastWord) {
	/* it's the last in the file, so it is also the last in the block */
	LastWord->WordPlace.Flags |= WPF_LASTINBLOCK;
	AddWord(LastWord);
    }

    if (SignalFlag > 1) {
	return;
    }

    if (AsciiTrace) {
	fprintf(stderr, "%s: %d: %s: indexed.\n",
	    progname,
	    FileInfo->FID,
	    FileInfo->Name
	);
    }
}
